/*
 * Copyright (c) 2015-2016 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                      Anthony Mallet on Mon Aug 24 2015
 */
#include "acpom.h"

#include <sys/time.h>

#include <cassert>

#include "pom_c_types.h"
#include "codels.h"


/* --- Task io ---------------------------------------------------------- */


/** Codel pom_io_start of task io.
 *
 * Triggered by pom_start.
 * Yields to pom_read.
 */
genom_event
pom_io_start(pom_ids *ids, genom_context self)
{
  struct timeval now;
  pom_snapshot s;

  ids->context = new pom_context_s;
  ids->history_length = 2;

  /* set initial state */
  gettimeofday(&now, NULL);

  ids->context->ts.sec = now.tv_sec;
  ids->context->ts.nsec = now.tv_usec * 1000;

  ids->context->state = ukf::state_s::Identity();
  ids->context->cov = ukf::state_s::dcov_s::Constant(0.);
  ids->context->cov.diagonal() = ukf::state_s::delta_s::Constant(0.1);

  return pom_read;
}


/** Codel pom_io_read of task io.
 *
 * Triggered by pom_read.
 * Yields to pom_pause_read, pom_insert.
 */
genom_event
pom_io_read(const pom_measure *measure, sequence_pom_portinfo_s *ports,
            sequence_or_pose_estimator_state *measurements,
            genom_context self)
{
  const or_pose_estimator_state *m;
  size_t i;

  assert(measurements->_length == 0);
  for(i = 0; i < ports->_length; i++) {
    /* get current port data */
    if (measure->read(ports->_buffer[i].name, self) != genom_ok)
      continue;
    m = measure->data(ports->_buffer[i].name, self);
    if (!m) continue;

    /* do not consider already read data */
    if (m->ts.nsec == ports->_buffer[i].last.nsec &&
        m->ts.sec == ports->_buffer[i].last.sec) continue;
    ports->_buffer[i].last = m->ts;

    /* record data */
    assert(measurements->_length < measurements->_maximum);
    measurements->_buffer[measurements->_length] = *m;
    measurements->_length++;
  }

  return measurements->_length ? pom_insert : pom_pause_read;
}


/** Codel pom_io_insert of task io.
 *
 * Triggered by pom_insert.
 * Yields to pom_pause_read.
 */
genom_event
pom_io_insert(sequence_or_pose_estimator_state *measurements,
              pom_context_s **context, uint16_t history_length,
              genom_context self)
{
  const or_pose_estimator_state *m;
  pom_snapshot s;
  struct timeval now;
  or_time_ts ts;
  size_t i;

  /* insert new measurements with an empty state */
  s.fused = false;

  for(i = 0; i < measurements->_length; i++) {
    m = &measurements->_buffer[i];
    s.measurement = *m;
    (*context)->history.insert(pom_history_s::value_type(m->ts, s));
  }
  measurements->_length = 0;

  /* truncate history */
  gettimeofday(&now, NULL);
  ts.sec = now.tv_sec - history_length;
  ts.nsec = now.tv_usec * 1000;

  (*context)->history.erase((*context)->history.upper_bound(ts),
                            (*context)->history.end());

  return pom_pause_read;
}


/** Codel pom_io_stop of task io.
 *
 * Triggered by pom_stop.
 * Yields to pom_ether.
 */
genom_event
pom_io_stop(pom_ids *ids, genom_context self)
{
  if (ids->context) delete ids->context;
  ids->context = NULL;

  return pom_ether;
}


/* --- Activity add_measurement ----------------------------------------- */

/** Codel pom_add_measurement of activity add_measurement.
 *
 * Triggered by pom_start.
 * Yields to pom_ether.
 * Throws pom_e_sys.
 */
genom_event
pom_add_measurement(const char port[128],
                    sequence_pom_portinfo_s *ports,
                    sequence_or_pose_estimator_state *measurements,
                    genom_context self)
{
  size_t i;

  for(i = 0; i < ports->_length; i++)
    if (!strcmp(ports->_buffer[i].name, port)) return pom_eexist(self);

  if (genom_sequence_reserve(measurements, ports->_length + 1))
    return pom_e_sys_error("add", self);
  if (genom_sequence_reserve(ports, ports->_length + 1))
    return pom_e_sys_error("add", self);

  strncpy(
    ports->_buffer[ports->_length].name, port, sizeof(ports->_buffer->name));
  ports->_buffer[ports->_length].last.sec = 0;
  ports->_buffer[ports->_length].last.nsec = 0;
  ports->_length++;

  return pom_ether;
}
